#   include	<stdio.h>
#   include	<string.h>
#   include	<iconv.h>
#   include	<ind.h>
#   include	<charnames.h>
#   include	<appDebugon.h>
#   include	"affix.h"

/************************************************************************/
/*									*/
/*  Make a finite automaton for 'Ted' spelling checking from an ispell	*/
/*  affix file and one ore more dictionary files.			*/
/*									*/
/*  USAGE: iindex <indFile> <affixFile> <dictionaries>			*/
/*		--encoding <enc> Gives the encoding of the affix file	*/
/*				and of the dictionary files. The	*/
/*				default is iso8859-1. For possible	*/
/*				values use 'iconv -l'.			*/
/*		<indFile>	Is the name of the automaton as it is	*/
/*				to be saved to disk.			*/
/*		<affixFile>	Is the name of an 'ispell' affix file.	*/
/*				The syntax of the rules is a little	*/
/*				simplified, but most files simply work.	*/
/*		<dictionaries>	One or more files with words, possibly	*/
/*				followed by a slash and applicable	*/
/*				affix rules. This is the usual 'ispell'	*/
/*				syntax.					*/
/*									*/
/*  1)  Check usage.							*/
/*  2)  Make an empty finite automaton.					*/
/*  3)  Open and parse the affix files.					*/
/*  4)  For all dictionaries....					*/
/*  5)  For all lines in the dictionary...				*/
/*  6)  Look for the end of the line, or the slash. Some hassle about	*/
/*	trailing spaces.						*/
/*  7)  If there is a slash, apply affix rules.				*/
/*  8)  Insert the word itself in the automaton.			*/
/*  9)  Obtain a minimal equivalent finite automation, discard the	*/
/*	original one. This may take quite a while. (At least several	*/
/*	minutes)							*/
/*  10) Renumber the states in the minimial automaton, this gives	*/
/*	better performance while spell checking due to better locality	*/
/*	of reference.							*/
/*  11) Write the so obtained automaton to file.			*/
/*  12) Ready, no cleanup as we stop anyway.				*/
/*									*/
/************************************************************************/

static int affixExtractdDictWord(	unsigned short *	word,
					int *			pLen,
					int *			pSlash,
					const AffixThrough *	at,
					char *			buf )
    {
    int				from= 0;
    int				past= 0;
    int				to= 0;

    int				l;

    l= strlen( buf );
    if  ( l > 0 )
	{ buf[l-1]= '\0';	}

    /*  6  */
    while( buf[from] && buf[from] != '/' )
	{
	int	c;
	int	cc;

	cc= buf[from];

	if  ( isspace( cc ) )
	    { c= cc; from++;	}
	else{
	    const int	lower= 0;

	    c= affixChar( at->atRules, buf, at->atIconv, lower, &from );

	    if  ( c < 0 )
		{ LDEB(c); return -1;	}
	    }

	word[to++]= c;

	if  ( ! isspace( cc ) )
	    { past= to;	}
	}

    word[past]= 0;

    *pLen= past;
    *pSlash= from;

    return 0;
    }

/************************************************************************/

static void affixGuessesForWord(	void *		voidind,
					const char *	word )
    {
    SpellCheckContext	scc;
    IndGuessList	igl;
    int			accepts;

    indInitSpellCheckContext( &scc );
    indInitGuessList( &igl );

    scc.sccStaticInd= voidind;

    /*
    scc.sccStaticInd= indMake();
    indPutUtf8( scc.sccStaticInd, "alleen" );
    */

    if  ( indGetUtf8( &accepts, voidind, word ) >=0 && accepts )
	{ appDebug( "ACCEPTED \"%s\"\n", word );	}
    else{ appDebug( "REJECTED \"%s\"\n", word );	}

    indCollectGuesses( &igl, &scc, word );

    appDebug( "%3d GUESSES:\n", igl.iglGuessCount );
    for ( accepts= 0; accepts < igl.iglGuessCount; accepts++ )
	{
	appDebug( "%3d: (%3d) \"%s\"\n", accepts,
				    igl.iglGuesses[accepts].igsScore,
				    igl.iglGuesses[accepts].igsWord );
	}

    scc.sccStaticInd= (void *)0;

    indCleanSpellCheckContext( &scc );
    indCleanGuessList( &igl );
    }

/************************************************************************/

# define BUL 300
static char buf[BUL+1];

int main(	int	argc,
		char **	argv )
    {
    int			rval= 0;
    FILE *		f= (FILE *)0;
    int			arg;

    void *		indmini;
    int			tn;
    unsigned short	one= 1;

    const int		useSuffixes= 0; /* Approach does not work ! */

    AffixThrough	at;

    affixInitThrough( &at );

    at.atEncodingIn= "ISO-8859-15";
    if  ( *(unsigned char *)&one )
	{ at.atEncodingOut= "UTF-16LE";	}
    else{ at.atEncodingOut= "UTF-16BE";	}

    if  ( argc > 2 && ! strcmp( argv[1], "--encoding" ) )
	{
	at.atEncodingIn= argv[2];
	argc -= 2;

	for ( arg= 1; arg < argc; arg++ )
	    { argv[arg]= argv[arg+2];	}
	}

    at.atIconv= iconv_open( at.atEncodingOut, at.atEncodingIn );
    if  (  at.atIconv == (iconv_t)-1 )
	{
	SSXDEB(at.atEncodingIn,at.atEncodingOut,at.atIconv);
	rval= 1; goto ready;
	}

    if  ( argc == 3 && ! strcmp( argv[1], "--Dump" ) )
	{
	at.atVoidInd= indRead( argv[2], 1 );
	if  ( ! at.atVoidInd )
	    { SXDEB(argv[1],at.atVoidInd); rval= 1; goto ready;	}

	indDump( at.atVoidInd );
	goto ready;
	}

    if  ( argc > 3 && ! strcmp( argv[1], "--Guess" ) )
	{
	at.atVoidInd= indRead( argv[2], 1 );
	if  ( ! at.atVoidInd )
	    { SXDEB(argv[1],at.atVoidInd); rval= 1; goto ready;	}

	for ( arg= 3; arg < argc; arg++ )
	    { affixGuessesForWord( at.atVoidInd, argv[arg] ); }

	goto ready;
	}

    /*  2  */
    at.atVoidInd= indMake();
    if  ( ! at.atVoidInd )
	{ XDEB(at.atVoidInd); rval= 1; goto ready;	}

    /*  2  */
    at.atSuffixInd= indMake();
    if  ( ! at.atSuffixInd )
	{ XDEB(at.atSuffixInd); rval= 1; goto ready;	}

    /*  3  */
    at.atRules= indAffixReadFile( argv[2], &at );
    if  ( ! at.atRules )
	{ SXDEB(argv[2],at.atRules); rval= 1; goto ready; }

    if  ( argc == 3 && ! strcmp( argv[1], "--List" ) )
	{
	indAffixListRules( at.atRules );
	goto ready;
	}

    if  ( useSuffixes )
	{
	if  ( affixCollectSuffixes( &at ) )
	    { LDEB(1); rval= 1; goto ready;	}
	}

    /*  1  */
    if  ( argc <= 3 )
	{ LLDEB(argc,3); rval= 1; goto ready;	}

    /*  4  */
    for ( arg= 3; arg < argc; arg++ )
	{
	int		line= 0;
	unsigned short	word[BUL+1];
	int		len= 0;
	int		slash= 0;

	const char *	filename= argv[arg];

	f= fopen( filename, "r" );
	if  ( ! f )
	    { SXDEB(filename,f); rval= 1; goto ready;	}

	/*  5  */
	line= 0;
	while( fgets( buf, BUL, f ) )
	    {
	    const int		prefixOnly= useSuffixes;

	    line++;

	    if  ( affixExtractdDictWord( word, &len, &slash, &at, buf ) < 0 )
		{ SLSDEB(filename,line,buf); rval= 1; goto ready; }

	    /*  7  */
	    if  ( buf[slash] )
		{
		if  ( indAffixApplyRules( at.atRules, prefixOnly, word, len,
						buf+ slash+ 1, at.atVoidInd ) )
		    { SLSDEB(filename,line,buf); rval= 1; goto ready; }
		}

	    /*  8  */
	    if  ( ( tn= indPutUtf16( at.atVoidInd, word ) ) < 0 )
		{ SLSDEB(filename,line,buf); rval= 1; goto ready; }
	    }

	if  ( useSuffixes )
	    {
	    rewind( f ); line= 0;

	    /*  5  */
	    line= 0;
	    while( fgets( buf, BUL, f ) )
		{
		const int		prefixOnly= 0;

		line++;

		if  ( affixExtractdDictWord( word, &len, &slash,
							    &at, buf ) < 0 )
		    { SLSDEB(filename,line,buf); rval= 1; goto ready; }

		/*  7  */
		if  ( buf[slash] )
		    {
		    if  ( indAffixApplyRules( at.atRules, prefixOnly, word, len,
						buf+ slash+ 1, at.atVoidInd ) )
			{ SLSDEB(filename,line,buf); rval= 1; goto ready; }
		    }
		}
	    }
	}

    /*  9  */
    indmini= indMini( at.atVoidInd );

    indFree( at.atVoidInd );
    at.atVoidInd= (void *)0;

    if  ( ! indmini )
	{ XDEB(indmini);; rval= 1; goto ready;	}

    if  ( indmini )
	{
	/*  10  */
	at.atVoidInd= indRenumber( indmini );

	if  ( ! at.atVoidInd )
	    { XDEB(at.atVoidInd);; rval= 1; goto ready; }
	indFree( indmini );
	}

    /*  11  */
    if  ( at.atVoidInd )
	{
	char *	indexName= argv[1];

	if  ( indWrite( at.atVoidInd, indexName ) )
	    { SDEB(indexName); rval= 1; goto ready; }
	}

  ready:
    if  ( f )
	{ fclose( f );	}

    affixCleanThrough( &at );

    /*  12  */
    return rval;
    }
